/*Prints all possible solutions to N-Queens problems.*/

#include <stdio.h>
#include <stdlib.h>
#include "mpi.h"

#define N 8                /*chess board dimension N*N*/

#define TREE_LEVEL_WORK 2  /*number of rows that the manager computes*/

#define NOQUEEN  0
#define QUEEN    1
#define ABOVE    0
#define LEFT    -1
#define RIGHT   +1

#define TRUE    1
#define FALSE   0

#define ALL MPI_COMM_WORLD
#define ANY MPI_ANY_SOURCE

#define TAG     27
#define BOARD   50
#define ROW     51
#define COL     52
#define REQ     53

#define on_board(row,col) (row>=0 && row<N && col>=0 && col<N)
#define last_row(row)     (row == N-1)
#define queen_in(row,col) (board[row][col]==QUEEN)

int  no_queen(int[][], int, int, int);
void initialize(int[][]);

/* Master */
#define limit_row(row) (row==TREE_LEVEL_WORK-1)
#define on_board_master(row,col) (row>=0 && row < TREE_LEVEL_WORK && col >=0 && col < N)
void place_queen_master(int[][], int, int);
void process_board(int[][], int, int);
int getFreeProc();

/* Worker */
void work();
void place_queen_worker(int[][], int, int);
void send_display_board(int[][]);

/* Display process */
void display_board();
void show_board(int[][]);

void main(int argc , char* argv[])
{
  int size, rank;	
  int procs, display;
  
  MPI_Init(&argc, &argv );  
  MPI_Comm_size(ALL,&procs);
  MPI_Comm_rank(ALL,&rank);

  size   = procs-1;  /* master plus number of workers */
  display= procs-1;  /* display process ID */

  if (rank==0)        /* Master Process*/
  {
    MPI_Status status;
    int board[N][N];
    int buffer;
    int src;
    int worker_cnt;
  
    initialize(board);
        	  
    /* generates work and distributes work to the free processors */
    place_queen_master(board, 0, 0);

    /* Terminate free processors */
    worker_cnt = size-1;      /* number of workers */
    while(worker_cnt>0)
    {
      MPI_Recv(&buffer, 1, MPI_INT, ANY, REQ, ALL, &status); 
      src= status.MPI_SOURCE; 
      buffer= -1;
      MPI_Send(&buffer, 1, MPI_INT, src, BOARD, ALL);
      worker_cnt--;
    }
  }
  else if(rank==display)     /* Display Process */
    display_board(); 
  else
    work();		     /* Worker Process */
  
  MPI_Finalize();	
}

/********************** Master ***************************************/
/* Recursively traverses the tree up to limit_row */
void place_queen_master(int board[][N], int row, int col)
{
  if(on_board_master(row,col))
  {
    if( no_queen(board,ABOVE,row,col) && no_queen(board,LEFT ,row,col) && no_queen(board,RIGHT,row,col) )
    {
      board[row][col] = QUEEN;

      if(limit_row(row)) 
	process_board(board, row+1, 0);

      place_queen_master(board,row+1,0);
      board[row][col] = NOQUEEN;
    }
    
    place_queen_master(board,row,col+1);
  }
}

/***Gets the free process no and ships the board, row and col to it***/
void process_board(int b[][N], int r, int c)
{
  int rank;
  int i,j;
  int buffer_size;
  int dest;
  
  buffer_size = N*N;
  MPI_Comm_rank(ALL, &rank);
  
  dest = getFreeProc();

  MPI_Send(b, buffer_size, MPI_INT, dest, BOARD, ALL);
  MPI_Send(&r, 1, MPI_INT, dest, ROW, ALL);
  MPI_Send(&c, 1, MPI_INT, dest, COL, ALL);
}

/*****A process is free if it has sent a REQ message to the master****/
int getFreeProc()
{
  MPI_Status status;
  int request;
  int src;
	
  MPI_Recv(&request, 1, MPI_INT, ANY, REQ, ALL, &status); 
  src = status.MPI_SOURCE;
  return(src);
}

/**************************** Worker *********************************/
/* Receives boards from the master,  expands them to the last level and 
   sends solution to the display process for displaying board on the 
   screen. As soon as the worker is free it sends a Request message to 
   the master.
*/
void work()
{
  int buffer[2];
  int rank;
  int buffer_size;
  int board[N][N];
  int row,col;
  int i,j;
  int done = 0;
  int size, procs, display;
  int TotalBoards=0;
  MPI_Status status;
  
  MPI_Comm_size(ALL,&procs);
  size = procs-1;
  display = procs-1;

  MPI_Comm_rank(ALL, &rank);
  buffer_size = N*N;
  if (!buffer)
    printf("Memory allocation error in proc : %d\n", rank);
  else
  {
    while (!done)
    {
      board[0][0]= -1;
      MPI_Send(buffer, 1, MPI_INT, 0, REQ, ALL); 
      MPI_Recv(board, buffer_size, MPI_INT, 0, BOARD, ALL, &status);
      
      if (board[0][0]!= -1)
      {
	TotalBoards++;
		  
	MPI_Recv(&row, 1, MPI_INT, 0, ROW, ALL, &status);      	      
	MPI_Recv(&col, 1, MPI_INT, 0, COL, ALL, &status);
	      	          
	place_queen_worker(board, row, col);	
      }
      else
        done = 1;
    }
   
    board[0][0]= -1;
    board[0][1]= TotalBoards;
    MPI_Send(board, buffer_size, MPI_INT, display, BOARD, ALL); 
  }
}

/************** Recursively computes the solution ********************/
void place_queen_worker(int board[][N], int row, int col)
{
  if(on_board(row,col))
  {
    if( no_queen(board,ABOVE,row,col) && no_queen(board,LEFT ,row,col) && no_queen(board,RIGHT,row,col) )
    {
      board[row][col] = QUEEN;
      if(last_row(row)) 
      send_display_board(board);
      place_queen_worker(board,row+1,0);
      board[row][col] = NOQUEEN;
    }
    
    place_queen_worker(board,row,col+1);
  }
}

/*************** sends the board to the display process ***************/
void send_display_board(int board[][N])
{
  int i, j;
  int buffer_size;
  int procs, display, size;
  
  MPI_Comm_size(ALL,&procs);
  size = procs-1;
  display = procs-1;

  buffer_size = N*N;
      
  MPI_Send(board, buffer_size, MPI_INT, display, BOARD, ALL);
}

/********************** General Routines ******************************/
int no_queen(int board[][N], int slant, int row, int col)
{
  int result;
  col += slant;
  row--;
  if(on_board(row,col))
    if(queen_in(row,col)) 
      result = FALSE;
    else 
      result = no_queen(board,slant,row,col);
  else
    result = TRUE;
  
  return result; 
}

/***************** initializing the board *****************************/
void initialize(int board[][N])
{
  int row,col,element;
  
  for(row=0;row<N;row++)
    for(col=0;col<N;col++)
      board[row][col] = NOQUEEN;
}



/* ************************** Display ******************************** */
/* Displays the board on screen */
void show_board(int board[][N])
{
  int row,col,element;
  for(row=0;row<N;row++)
  {
    for(col=0;col<N;col++)
      if(board[row][col] == QUEEN)
        printf("Q ");
      else
        printf("- ");

     printf("\n");
  }
  printf("\n\n");
}

/* Receives boards from workers and displays them.
   Terminates only after all the workers' boards have been displayed
*/
void display_board()
{
  MPI_Status status;
  int buffer_size;
  int brd[N][N];
  int cnt= 0;
  int i,j;
  
  int src;
  int done = 0;
  int procs, display, size;
  int worker_cnt;
  
  MPI_Comm_size(ALL,&procs);
  size = procs-1;
  display = procs-1;
  worker_cnt = size-1;

  buffer_size = N*N;
    
  while(!done)
  {
    MPI_Recv(brd, buffer_size, MPI_INT, ANY, BOARD, ALL, &status); 
    src = status.MPI_SOURCE;
    if (brd[0][0]==-1)
    {
      worker_cnt --;
      if (worker_cnt <= 0)
      done = 1;
    }
    else
    {
      cnt++;
      printf("\nSolution# %i (from Proccess %i):\n", cnt, src);      
      show_board(brd); 
    }
  }
}

